# -*- coding: utf-8 -*-
#
#  Copyright (C) 2001, 2002 by Tamito KAJIYAMA
#  Copyright (C) 2002, 2003 by MATSUMURA Namihiko <nie@counterghost.net>
#  Copyright (C) 2002-2009 by Shyouzou Sugitani <shy@users.sourceforge.jp>
#  Copyright (C) 2003 by Shun-ichi TAHARA <jado@flowernet.gr.jp>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License (version 2) as
#  published by the Free Software Foundation.  It is distributed in the
#  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the GNU General Public License for more details.
#

import os
import sys
import time
import StringIO
import codecs

if 'DISPLAY' in os.environ:
    import gtk ## FIXME
    import gobject

import ninix.dll
import ninix.makoto
import ninix.update
import ninix.surface
import ninix.balloon
import ninix.sakura
import ninix.version


class Ghost:

    def __init__(self, app, data, default_path, communicate, debug=0):
        self.app = app
        self.char = 2 # 'sakura' and 'kero'
        self.__sender = 'ninix-aya'
        self.__sakura  = ninix.sakura.Sakura(self, debug)
        self.keep_silence(False)
        self.__observer = []
        self.event_queue = []
        saori_lib = ninix.dll.Library(
            'saori', default_path, ghost=self)
        self.__dll = ninix.dll.Library(
            'shiori', default_path, saori_lib=saori_lib)
        self.debug = debug
        self.__surface = ninix.surface.Surface(self, debug)
        self.__sakura.set_surface(self.__surface)
        self.__balloon = ninix.balloon.Balloon(self, debug)
        self.__sakura.set_balloon(self.__balloon)
        self.communicate = communicate
        # create vanish dialog
        self.__vanish_dialog = VanishDialog(self)
        self.__running = 0
        self.__temp_mode = 0
        self.__charset = 'Shift_JIS'
        self.__cantalk = True
        self.top_margin = 0
        self.bottom_margin = 0
        desc, shiori_dir, use_makoto, surface_set, balloon, prefix, \
              shiori_dll, shiori_name = data
        self.new(desc, shiori_dir, use_makoto, surface_set, balloon, prefix,
                 shiori_dll, shiori_name) ## FIXME

    ### FIXME ###
    def enqueue_script(self, script, sender, handle,
                       host, show_sstp_marker, use_translator,
                       db=None, request_handler=None):
        if not self.__sakura.script_queue and \
           not self.__sakura.time_critical_session and not self.__sakura.passivemode:
            if self.__sakura.sstp_request_handler:
                self.__sakura.sstp_request_handler.send_sstp_break()
                self.__sakura.sstp_request_handler = None
            self.__sakura.reset_script(1)
        self.__sakura.script_queue.append((script, sender, handle, host,
                                  show_sstp_marker, use_translator,
                                  db, request_handler))

    reset_event = ['OnGhostChanging', 'OnShellChanging', 'OnVanishSelected',
                   'OnNinixReloading'] ## FIXME

    def enqueue_event(self, event, *arglist, **argdict): ## FIXME
        for key in argdict:
            assert key in ['proc'] # trap typo, etc.
        if event in self.reset_event:
            self.__sakura.reset_script(1)
        self.event_queue.append((event, arglist, argdict))

    EVENT_SCRIPTS = {
        'OnUpdateBegin': \
        ''.join((r'\t\h\s[0]',
                 unicode(_('Network Update has begun.'), 'utf-8'),
                 r'\e')),
        'OnUpdateComplete': \
        ''.join((r'\t\h\s[5]',
                 unicode(_('Network Update completed successfully.'), 'utf-8'),
                 r'\e')),
        'OnUpdateFailure': \
        ''.join((r'\t\h\s[4]',
                 unicode(_('Network Update failed.'), 'utf-8'),
                 r'\e')),
        }

    def handle_event(self): ## FIXME
        while self.event_queue:
            event, arglist, argdict = self.event_queue.pop(0)
            proc = argdict.get('proc', None)
            argdict = {'default': self.EVENT_SCRIPTS.get(event)}
            if self.__sakura.notify_event(event, *arglist, **argdict):
                if proc is not None:
                    self.__sakura.script_post_proc.append(proc)
                return 1
            elif proc is not None:
                proc()
                return 1
        return 0

    def get_callback(self, name):
        callback = {'close all':        self.app.close_all_ghosts,
                    'select sakura':    self.app.select_sakura,
                    'start':            self.app.start_sakura_cb,
                    'select plugin':    self.app.select_plugin,
                    'select nekodorif': self.app.select_nekodorif,
                    'select kinoko':    self.app.select_kinoko,
                    }
        if name in callback:
            return callback[name]

    def notify_surface_click(self, button, click, side, x, y): ## FIXME ## FIXME: sakura
        self.notify_observer('raise', (side)) ## FIXME: automagical raise 
        if button == 1 and \
           self.__sakura.mouse_button1 == ninix.sakura.BUTTON1_RAISE or \
           button == 3 and \
           self.__sakura.mouse_button3 == ninix.sakura.BUTTON3_RAISE:
            self.__sakura.raise_all()
        elif button == 1 and \
             self.__sakura.mouse_button1 == ninix.sakura.BUTTON1_LOWER or \
             button == 3 and \
             self.__sakura.mouse_button3 == ninix.sakura.BUTTON3_LOWER:
            self.__sakura.lower_all()
        if self.__sakura.vanished:
            if side == 0 and button == 1:
                if self.__sakura.sstp_request_handler:
                    self.__sakura.sstp_request_handler.send_sstp_break()
                    self.__sakura.sstp_request_handler = None
                self.__sakura.reset_script(1)
                self.__sakura.notify_event('OnVanishButtonHold', default=r'\e')
                self.__sakura.vanished = 0
            return
        if self.updateman.is_active():
            if button == 1 and click == 2:
                self.updateman.interrupt()
            return
        if self.__sakura.time_critical_session:
            return
        elif button in [1, 3] and click == 1:
            if self.__sakura.passivemode and \
               self.__sakura.processed_script is not None:
                return
            part = self.__surface.get_touched_region(side, x, y)
            self.__sakura.notify_event('OnMouseClick',
                                       x, y, 0, side, part, button)
        elif self.__sakura.passivemode:
            return
        elif button == 1 and click == 2:
            if self.__sakura.sstp_request_handler:
                self.__sakura.sstp_request_handler.send_sstp_break()
                self.__sakura.sstp_request_handler = None
            part = self.__surface.get_touched_region(side, x, y)
            self.__sakura.notify_event('OnMouseDoubleClick',
                                       x, y, '', side, part)

    def notify_surface_mouse_motion(self, side, x, y, part):
        if self.__sakura.surface_mouse_motion is not None:
            return
        if part:
            self.__sakura.surface_mouse_motion = (side, x, y, part)
        else:
            self.__sakura.surface_mouse_motion = None

    def notify_balloon_click(self, button, click, side): ## FIXME
        self.__sakura.notify_balloon_click(button, click, side)

    def notify_link_selection(self, link_id, text, number): ## FIXME
        self.__sakura.notify_link_selection(link_id, text, number)

    def notify_user_teach(self, data): ## FIXME
        self.__sakura.notify_user_teach(data)

    def busy(self):
        if self.__sakura.busy() or self.updateman.is_active():
            return True
        else:
            return False

    def sakura_is_running(self):
        return self.__sakura.running

    def cantalk(self):
        if self.__cantalk:
            return True
        else:
            return False

    def is_running(self):
        if self.__running:
            return 1
        else:
            return 0

    def set_observer(self, observer):
        if observer not in self.__observer:
            self.__observer.append(observer)

    def notify_observer(self, event, args=()):
        for observer in self.__observer:
            observer.notify_observer(event, args)

    def delete_observer(self, observer):
        for i in range(len(self.__observer)):
            if self.__observer[i] == observer:
                del self.__observer[i]
                break

    def raise_surface(self, side): ## FIXME: sakura
        self.__surface.raise_(side)

    def lower_surface(self, side): ## FIXME: sakura
        self.__surface.lower(side)

    def get_kinoko_position(self, baseposition):
        return self.__sakura.get_kinoko_position(baseposition)

    def get_surface_size(self, side):
        return self.__sakura.get_surface_size(side)

    def get_surface_position(self, side): ## FIXME: sakura
        result = self.__surface.get_position(side)
        if result is None:
            result = (0, 0)
        return result

    def position_balloons(self): ## FIXME
        self.__sakura.position_balloons()


    def get_balloon_position(self, side): ## FIXME: sakura
        result = self.__balloon.get_position(side)
        if result is None:
            result = (0, 0)
        return result

    def get_balloon_size(self, side): ## FIXME: sakura
        result = self.__balloon.get_balloon_size(side)
        if result is None:
            result = (0, 0)
        return result

    def set_surface_position(self, side, x, y):
        self.__sakura.set_surface_position(side, x, y)

    def get_sakura(self): ## FIXME: this method should be removed
        return self.__sakura

    def is_paused(self):
        return self.__sakura.script_mode == self.__sakura.PAUSE_MODE

    def is_talking(self):
        if self.__sakura.processed_script or \
           self.__sakura.processed_text:
            return 1
        else:
            return 0

    def set_use_pna(self, flag): ## FIXME: sakura
        self.__surface.set_use_pna(flag)
        self.__balloon.set_use_pna(flag)

    def set_surface_alpha(self, alpha): ## FIXME: sakura
        self.__surface.set_alpha_channel(alpha)

    def set_balloon_alpha(self, alpha): ## FIXME: sakura
        self.__balloon.set_alpha_channel(alpha)

    def set_sink_after_talk(self, flag):
        self.__sakura.set_sink_after_talk(flag)

    def set_raise_before_talk(self, flag):
        self.__sakura.set_raise_before_talk(flag)

    def set_animation_quality(self, quality): ## FIXME: sakura
        self.__surface.set_animation_quality(quality)

    def set_seriko_inactive(self, flag): ## FIXME: sakura
        self.__surface.set_seriko_inactive(flag)

    def set_balloon_position(self, side, x, y): ## FIXME ## FIXME: sakura
        self.__balloon.set_position(side, x, y)

    def set_balloon_direction(self, side, direction): ## FIXME: sakura
        self.__balloon.set_direction(side, direction)

    def set_balloon_scalling(self, flag): ## FIXME: sakura
        self.__balloon.set_scalling(flag)

    def get_balloon_scalling(self): ## FIXME: sakura
        return self.__balloon.get_scalling()

    def get_helpers(self):
        return self.__sakura.helpers

    def balloon_is_shown(self, side): ## FIXME: sakura
        if self.__balloon and self.__balloon.is_shown(side):
            return 1
        else:
            return 0

    def surface_is_shown(self, side): ## FIXME: sakura
        if self.__surface and self.__surface.is_shown(side):
            return 1
        else:
            return 0

    def set_debug(self, debug): ## FIXME ## FIXME: sakura
        self.debug = debug
        self.__surface.set_debug(debug)

    def get_top_margin(self):
        return self.top_margin

    def set_top_margin(self, margin):
        self.top_margin = margin
        self.__sakura.align_current()
        self.notify_observer('set top margin')

    def get_bottom_margin(self):
        return self.bottom_margin

    def set_bottom_margin(self, margin):
        self.bottom_margin = margin
        self.__sakura.align_current()
        self.notify_observer('set bottom margin')

    def set_browser(self, command):
        self.__sakura.set_browser(command)

    def set_helpers(self, list):
        self.__sakura.set_helpers(list)

    def set_script_speed(self, speed):
        self.__sakura.set_script_speed(speed)

    def get_script_speed(self):
        return self.__sakura.script_speed

    def set_event_kill_list(self, list):
        self.__sakura.set_event_kill_list(list)

    def set_surface_scale(self, scale):
        self.__sakura.set_surface_scale(scale)
        self.notify_observer('set scale')

    def get_surface_scale(self):
        return self.__sakura.get_surface_scale()

    def stop(self):
        if not self.__running:
            return
        self.notify_observer('finalize')
        self.__running = 0
        self.save_history()
        self.communicate.rebuild_ghostdb(self, None)
        self.__sakura.hide_all()
        self.__sakura.stop()
        self.__surface.finalize()
        self.__balloon.finalize()

    def restart(self):
        self.load_history()
        self.__sakura.restart()

    def set_balloon(self, desc, balloon):
        self.__balloon.new(desc, balloon)
        for side in range(2, self.char):
            self.__balloon.add_window(side)
        for side in range(self.char):
            balloon_win = self.__balloon.get_window(side)
            surface_win = self.__surface.get_window(side)
            balloon_win.set_transient_for(surface_win)

    def set_surface(self, desc, alias, surface, name, dir, tooltips):
        self.__surface.new(desc, alias, surface, name, dir, tooltips)
        for side in range(2, self.char):
            default = self.desc.get('char%d.seriko.defaultsurface' % side)
            self.__surface.add_window(side, default)
        icon = self.desc.get('icon', None)
        if icon is not None:
            icon_path = os.path.join(self.shiori_dir, icon)
            if not os.path.exists(icon_path):
                icon_path = None
        else:
            icon_path = None
        self.__surface.set_icon(icon_path)

    def set_balloon_fonts(self, fonts):
        self.__balloon.set_balloon_fonts(fonts)

    def get_balloon_fonts(self):
        return self.__balloon.get_balloon_fonts()

    def set_mouse_button1(self, button1):
        self.__sakura.set_mouse_button1(button1)

    def set_mouse_button3(self, button3):
        self.__sakura.set_mouse_button3(button3)

    def finalize(self):
        if not self.__temp_mode:
            self.unload_shiori()
        self.stop()

    def get_vanished_count(self):
        return self.vanished_count

    def set_vanished_count(self, count):
        self.vanished_count = count

    def vanish_by_myself(self):
        count = self.get_vanished_count()
        self.set_vanished_count(count + 1)
        self.set_ghost_time(0)
        self.app.stop_sakura(self.app.vanish_sakura)

    def get_ghost_time(self):
        return self.ghost_time

    def set_ghost_time(self, clock):
        self.ghost_time = clock

    def increment_ghost_time(self):
        if not self.__temp_mode:
            self.set_ghost_time(self.get_ghost_time() + 1)

    def unload_shiori(self):
        self.shiori.unload()

    def set_cantalk(self, flag):
        if flag:
            self.__cantalk = True
        else:
            self.__cantalk = False
        self.app.select_current_sakura()

    def notify_iconified(self):
        self.set_cantalk(False)
        if not self.__sakura.passivemode:
            self.__sakura.reset_script(1)
            self.__sakura.stand_by(1)
            self.__sakura.notify_event('OnWindowStateMinimize')
        self.notify_observer('iconified')

    def notify_deiconified(self):
        if not self.cantalk():
            self.set_cantalk(True)
            if not self.__sakura.passivemode:
                self.__sakura.notify_event('OnWindowStateRestore')
        self.notify_observer('deiconified')

    def notify_event(self, event, *arglist, **argdict): ## FIXME
        self.__sakura.notify_event(event, *arglist, **argdict)

    def get_own_balloon_name(self):
        if self.own_balloon:
            desc, balloon = self.own_balloon
            name = desc.get('name', unicode(_('Own Balloon'), 'utf-8'))
        else:
            name = None
        return name

    def get_current_balloon_name(self):
        return self.__balloon.get_balloon_name()

    def vanish(self):
        if self.busy():
            gtk.gdk.beep() ## FIXME
            return
        self.__sakura.notify_event('OnVanishSelecting')
        self.__vanish_dialog.show()

    def notify_vanish_confirmation(self):
        def proc(self=self):
            count = self.get_vanished_count()
            self.set_vanished_count(count + 1)
            self.set_ghost_time(0)
            gobject.idle_add(self.app.vanish_sakura, self) ## FIXME
        self.enqueue_event('OnVanishSelected', proc=proc)
        self.__sakura.vanished = 1 ## FIXME

    def load_shiori(self): ##FIXME
        if self.shiori and self.shiori.load(self.shiori_dir):
            if getattr(self.shiori, 'show_description', None):
                self.shiori.show_description()
        else:
            sys.stderr.write('%s cannot load SHIORI(%s)\n' % \
                             (self.get_selfname(), self.shiori_name))
        self.__charset = 'Shift_JIS' # default
        charset = self.get_event_response('charset')
        if charset:
            try:
                codecs.lookup(charset)
            except:
                sys.stderr.write('Unsupported charset %s' % repr(charset))
            else:
                self.__charset = charset
        else:
            pass
        self.get_event_response('basewareversion',
                                ninix.version.VERSION,
                                'ninix-aya',
                                ninix.version.NUMBER,
                                event_type='NOTIFY')

    def new(self, desc, shiori_dir, use_makoto, surface_set, balloon, prefix,
            shiori_dll, shiori_name): ## FIXME
        self.shiori = None
        self.desc = desc
        self.shiori_dir = shiori_dir
        self.use_makoto = use_makoto
        self.surface_set = surface_set
        self.own_balloon = balloon
        self.prefix = prefix
        self.shiori_dll = shiori_dll
        self.shiori_name = shiori_name
        name = (shiori_dll, shiori_name)
        self.shiori = self.__dll.request(name)
        self.updateman = ninix.update.NetworkUpdate(self)
        char = 2
        while self.desc.get('char%d.seriko.defaultsurface' % char) is not None:
            char += 1
        if char > 2:
            self.char = char
        # XXX
        if self.desc.get('name') == 'BTH小っちゃいってことは便利だねっ':
            self.set_SSP_mode(1)
        else:
            self.set_SSP_mode(0)

    def notify_start(self, init, vanished, ghost_changed, name, path):
        if self.__temp_mode:
            default = None
        else:
            default = ninix.version.VERSION_INFO
        if init:
            if self.get_ghost_time() == 0:
                if not self.__sakura.notify_event('OnFirstBoot',
                                                  self.get_vanished_count()):
                    self.__sakura.notify_event('OnBoot', default=default)
            else:
                self.__sakura.notify_event('OnBoot', default=default)
            self.__sakura.notify_event('OnDisplayChange',
                                       gtk.gdk.visual_get_best_depth(),
                                       gtk.gdk.screen_width(),
                                       gtk.gdk.screen_height(),
                                       event_type='NOTIFY')
        elif vanished:
            if self.get_ghost_time() == 0:
                if self.__sakura.notify_event('OnFirstBoot',
                                              self.get_vanished_count()):
                    return
            elif self.__sakura.notify_event('OnVanished', name):
                return
            self.__sakura.notify_event('OnBoot', default=default)
        elif ghost_changed:
            if self.get_ghost_time() == 0:
                if self.__sakura.notify_event('OnFirstBoot',
                                              self.get_vanished_count()):
                    return
            elif self.__sakura.notify_event('OnGhostChanged', name):
                return
            self.__sakura.notify_event('OnBoot', default=default)
        else:
            self.__sakura.notify_event('OnShellChanged', name, name, path)

    def start(self, item, init, temp, vanished, ghost_changed, ignore_default,
              prev_name):
        if self.is_running():
            if temp:
                self.enter_temp_mode()
            else:
                if self.__temp_mode == 1:
                    self.__temp_mode = 2
                    self.load_shiori()
                    self.notify_start(init, vanished, ghost_changed,
                                      prev_name, '')
            return
        self.set_ghost_time(0)
        self.set_vanished_count(0)
        self.__running = 1
        self.__temp_mode = temp
        self.current = item
        set_type, i, j = item
        if set_type == 'g':
            print 'ghost', item
            if not self.surface_set:
                shell_name, surface_set, balloon = self.app.request_shell(0)
                name, surface_dir, surface_desc, surface_alias, surface, surface_tooltips = \
                      surface_set[0]
                surface_set = None
            else:
                name, surface_dir, surface_desc, surface_alias, surface, surface_tooltips = \
                      self.surface_set[j]
                surface_set = self.surface_set
                balloon = self.own_balloon
        elif set_type == 's':
            print 'shell', item
            shell_name, surface_set, balloon = self.app.request_shell(i)
            name, surface_dir, surface_desc, surface_alias, surface, surface_tooltips = \
                  surface_set[j]
        if not surface_set:
            surface_name = None
        elif len(surface_set) == 1:
            surface_name = unicode(_('Master'), 'utf-8')
        else:
            surface_name = name
        if ghost_changed:
            name = prev_name
        self.set_surface(surface_desc, surface_alias, surface, surface_name,
                         surface_dir, surface_tooltips)
        if ignore_default or not balloon:
            balloon_name = self.desc.get('balloon', '')
            number = None
            if balloon_name:
                number = self.app.find_balloon_by_name(balloon_name)
            if number is None:
                default_balloon = self.app.get_default_balloon()
                number = self.app.find_balloon_by_name(default_balloon) or 0
            balloon = self.app.request_balloon(number)
        desc, balloon = balloon
        self.set_balloon(desc, balloon)
        if not temp:
            self.load_shiori()
        self.restart()
        self.__sakura.start()
        self.notify_start(init, vanished, ghost_changed, name, surface_dir)
        self.set_timeout()

    def save_history(self):
        path = os.path.join(self.get_prefix(), 'HISTORY')
        try:
            f = open(path, 'w')
        except IOError, (code, message):
            sys.stderr.write('cannot write %s\n' % path)
        else:
            time = self.get_ghost_time()
            vanished_count = self.get_vanished_count()
            f.write('time, %s\n' % time)
            f.write('vanished_count, %s\n' % vanished_count)
            f.close()

    def load_history(self):
        path = os.path.join(self.get_prefix(), 'HISTORY')
        if os.path.exists(path):
            ghost_time = 0
            ghost_vanished_count = 0
            try:
                f = open(path, 'r')
            except IOError, (code, message):
                sys.stderr.write('cannot read %s\n' % path)
            else:
                for line in f:
                    if ',' not in line:
                        continue
                    key, value = [x.strip() for x in line.split(',', 1)]
                    if key == 'time':
                        try:
                            ghost_time = int(value)
                        except:
                            pass
                    elif key == 'vanished_count':
                        try:
                            ghost_vanished_count = int(value)
                        except:
                            pass
                f.close()
            self.set_ghost_time(ghost_time)
            self.set_vanished_count(ghost_vanished_count)
        else:
            self.set_ghost_time(0)
            self.set_vanished_count(0)

    def get_prefix(self):
        return self.prefix

    def translate(self, s):
        if s is not None:
            if self.use_makoto:
                s = ninix.makoto.execute(s)
            else:
                r = self.get_event_response('OnTranslate', s, translate=0)
                if r:
                    s = r
        return s

    def getaistringrandom(self): # obsolete
        result = self.get_event_response('OnAITalk')
        return self.translate(result)

    def getdms(self):
        result = self.get_event_response('dms')
        return self.translate(result)

    def getword(self, word_type):
        result = self.get_event_response('%s' % word_type)
        return self.translate(result)

    def getstring(self, name):
        return self.get_event_response(name)

    def get_name(self):
        return self.desc.get('name', unicode(_('Sakura&Unyuu'), 'utf-8'))

    def get_username(self):
        return self.getstring('username') or \
               self.__surface.get_username() or \
               self.desc.get('user.defaultname', unicode(_('User'), 'utf-8'))

    def get_selfname(self):
        return self.__surface.get_selfname() or \
               self.desc.get('sakura.name', unicode(_('Sakura'), 'utf-8'))

    def get_selfname2(self):
        return self.__surface.get_selfname2() or \
               self.desc.get('sakura.name2', unicode(_('Sakura'), 'utf-8'))

    def get_keroname(self):
        return self.__surface.get_keroname() or \
               self.desc.get('kero.name', unicode(_('Unyuu'), 'utf-8'))

    def get_friendname(self):
        return self.__surface.get_friendname() or \
               self.desc.get('sakura.friend.name',
                             unicode(_('Tomoyo'), 'utf-8'))

    def set_SSP_mode(self, flag): # XXX
        if flag:
            self.__sender = 'SSP'
        else:
            self.__sender = 'ninix-aya'

    def get_event_response_with_communication(self, event, *arglist, **argdict):
        if self.__temp_mode == 1:
            return ''
        for key in argdict:
            assert key in ['event_type', 'translate'] # trap typo, etc.
        ref = arglist
        event_type = argdict.get('event_type', 'GET')
        translate = argdict.get('translate', 1)
        header = ''.join(('%s SHIORI/3.0\r\n' % event_type,
                          'ID: %s\r\n' % event,
                          'Sender: %s\r\n' % self.__sender,
                          'SecurityLevel: local\r\n'))
        for i in range(len(ref)):
            if ref[i] is not None:
                header = ''.join((header,
                                  'Reference', str(i), ': ',
                                  str(ref[i]), '\r\n'))
        header = ''.join((header, '\r\n'))
        header = header.encode(self.__charset, 'ignore')
        response = self.shiori.request(header)
        if event_type != 'NOTIFY' and self.cantalk():
            result, to = self.get_value(response)
            if translate:
                result = self.translate(result)
        else:
            result, to = '', None
        if self.__temp_mode == 2 and event != 'charset':
            result, to = '', None
        if result is None:
            result = ''
        if to and result:
            def proc():
                self.communicate.send_message(
                    to, self.__sakura.get_selfname(), result)
            communication = proc
        else:
            communication = None
        return result, communication

    def get_event_response(self, event, *arglist, **argdict):
        result, communication = self.get_event_response_with_communication(event, *arglist, **argdict)
        return result

    def get_value(self, response): # FIXME: check return code
        header = StringIO.StringIO(response)
        result = {}
        to = None
        for line in header:
            line = line.strip()
            if not line:
                continue
            if ':' not in line:
                continue
            key, result[key] = [x.strip() for x in line.split(':', 1)]
        for key in result.keys():
            result[key] = unicode(result[key], self.__charset, 'ignore')
        if 'Reference0' in result:
            to = result['Reference0']
        if 'Value' in result:
            return result['Value'], to
        else:
            return None, to

    def teach(self, word):
        result = self.get_event_response('OnTeach', word)
        return self.translate(result)

    def update(self):
        if self.updateman.is_active():
            return
        homeurl = self.getstring('homeurl')
        if not homeurl:
            self.__sakura.start_script(
                ''.join((r'\t\h\s[0]',
                         unicode(_("I'm afraid I don't have Network Update yet."), 'utf-8'),
                         r'\e')))
            self.__sakura.balloon.hide_sstp_message()
            return
        ghostdir = self.get_prefix()
        print 'homeurl =', homeurl
        print 'ghostdir =', ghostdir
        self.updateman.start(homeurl, ghostdir)

    def quit(self):
        self.app.stop_sakura(self)

    def edit_preferences(self):
        self.app.edit_preferences()

    def open_ghost_manager(self):
        self.app.open_ghost_manager()

    def select_ghost(self, sequential, event=1):
        self.app.select_ghost(self, sequential, event)

    def select_ghost_by_name(self, name, event=1):
        self.app.select_ghost_by_name(self, name, event)

    def select_shell_by_name(self, name, event=1):
        self.app.select_shell_by_name(self, name, event)

    def select_balloon(self, widget, item):
        if not widget.active:
            return
        if self.own_balloon:
            desc, balloon = self.own_balloon
            name = self.get_own_balloon_name()
        else:
            name = None
        if item != name:
            desc, balloon = self.app.get_balloon_description(item)
        self.__balloon.hide_all()
        self.set_balloon(desc, balloon)
        self.__sakura.set_balloon(self.__balloon)
        self.__balloon.set_balloon_default()
        print 'balloon', item

    def enter_temp_mode(self):
        if not self.__temp_mode:
            self.__temp_mode = 2

    def leave_temp_mode(self):
        self.__temp_mode = 0

    def get_silent_time(self):
        return self.silent_time

    def keep_silence(self, quiet):
        if quiet:
            self.silent_time = time.time()
        else:
            self.silent_time = 0
            self.reset_idle_time()

    def get_idle_time(self):
        now = time.time()
        idle = now - self.idle_start
        return idle

    def reset_idle_time(self):
        self.idle_start = time.time()

    ###   TIMEOUT   ###
    def set_timeout(self): ## FIXME
        gobject.timeout_add(10, self.do_idle_tasks) # 10ms

    reload_event = None

    def do_idle_tasks(self):
        if not self.__running:
            return False
        if self.__temp_mode:
            self.__sakura.process_script()
            if not self.busy() and \
               not self.__sakura.script_queue and \
               not (self.__sakura.processed_script or \
                    self.__sakura.processed_text):
                if self.__temp_mode == 1:
                    time.sleep(1.4)
                    self.finalize()
                    self.app.close_ghost(self) ## FIXME
                    self.app.reset_sstp_flag() ## FIXME
                    return False
                else:
                    self.app.reset_sstp_flag() ## FIXME                    
                    self.leave_temp_mode()
                    return True
            else:
                return True
        if self.reload_event and not self.busy() and \
           not (self.__sakura.processed_script or \
                self.__sakura.processed_text):
            self.__sakura.hide_all()
            sys.stdout.write('reloading....\n')
            self.shiori.unload() ## FIXME
            self.app.stop_sakura(self, self.app.reload_current_sakura, self) ## FIXME
            self.restart()
            sys.stdout.write('done.\n')
            self.enqueue_event(*self.reload_event)
            self.reload_event = None
        # continue network update (enqueue events)
        if self.updateman.is_active():
            self.updateman.run()
            while 1:
                event = self.updateman.get_event()
                if not event:
                    break
                if event[0] == 'OnUpdateComplete' and event[1] == 'changed':
                    self.reload_event = event
                else:
                    self.enqueue_event(*event)
        self.__sakura.process_script()
        return True

    def launch_browser(self, event, args): ## FIXME
        title, url = args
        if self.__sakura.is_URL(url):
            self.__sakura.launch_browser(url)
        self.enqueue_event('OnRecommandedSiteChoice', title, url)

    def close(self):
        if self.busy():
            gtk.gdk.beep() ## FIXME
            return
        self.__sakura.reset_script(1)
        self.enqueue_event('OnClose')

    def about(self):
        if self.busy():
            gtk.gdk.beep() ## FIXME
            return
        self.__sakura.start_script(ninix.version.VERSION_INFO)

    def network_update(self):
        if self.busy():
            gtk.gdk.beep() ## FIXME
            return
        self.update()

    def reload(self): ## FIXME
        self.app.reload(self)

    def show_usage(self):
        self.app.show_usage()

class VanishDialog:

    def __init__(self, ghost):
        self.__ghost = ghost
        self.window = gtk.Dialog()
        self.window.connect('delete_event', self.cancel)
        self.window.set_title('Vanish')
        self.window.set_modal(True)
        self.window.set_position(gtk.WIN_POS_CENTER)
        self.label = gtk.Label(unicode(_('Vanish'), 'utf-8'))
        self.window.vbox.pack_start(self.label, padding=10)
        self.label.show()
        box = gtk.HButtonBox()
        box.set_layout(gtk.BUTTONBOX_END)
        self.window.action_area.pack_start(box)
        box.show()
        button = gtk.Button(unicode(_('Yes'), 'utf-8'))
        button.connect('clicked', self.ok)
        box.add(button)
        button.show()
        button = gtk.Button(unicode(_('No'), 'utf-8'))
        button.connect('clicked', self.cancel)
        box.add(button)
        button.show()

    def set_message(self, message):
        self.label.set_text(message)

    def show(self):
        self.window.show()

    def ok(self, widget, event=None):
        self.window.hide()
        self.__ghost.notify_vanish_confirmation()
        return True

    def cancel(self, widget, event=None):
        self.window.hide()
        self.__ghost.notify_event('OnVanishCancel')
        return True


def test():
    pass

if __name__ == '__main__':
    test()
